mod dirs;


use std::ffi::OsStr;
use std::fs;
use std::io::{Read, Write};
use std::path::PathBuf;

use anyhow::{bail, ensure, Result};
use clap::{Parser, Subcommand};


#[derive(Debug, Parser)]
pub struct Args {
    #[command(subcommand)]
    cmd: Command,
}

#[derive(Debug, Subcommand)]
enum Command {
    /// Add a project
    Add { path: PathBuf },

    /// List existing projects
    List,
}


#[derive(Debug, Default)]
struct Projects {
    list: Vec<PathBuf>,
}


fn main() -> Result<()> {
    let args = Args::parse();

    let mut projects = read_projects_file()?;

    let result = match &args.cmd {
        Command::Add { path } => add_project(&mut projects, path),
        Command::List => list_projects(&projects),
    };

    if let Err(err) = result {
        eprintln!("{err}");
        std::process::exit(1);
    }

    Ok(())
}


fn add_project<P>(projects: &mut Projects, path: P) -> Result<()>
where
    P: Into<PathBuf>,
{
    let path = path.into();

    ensure!(path.is_dir(), "Project path does not exists");
    ensure!(
        !projects.list.contains(&path.to_path_buf()),
        "Project path already registered"
    );

    projects.list.push(path);
    write_projects_file(projects)?;

    Ok(())
}


fn list_projects(projects: &Projects) -> Result<()> {
    for project in &projects.list {
        println!("{project:?}")
    }

    Ok(())
}

fn write_projects_file(projects: &Projects) -> Result<()> {
    let projects_file = dirs::data_path().join("projects");

    let mut file = fs::OpenOptions::new()
        .write(true)
        .create(true)
        .truncate(true)
        .open(projects_file)?;

    for project_path in &projects.list {
        let os_str = project_path.as_os_str();
        let bytes = os_str.as_encoded_bytes();
        file.write_all(bytes)?;
        file.write_all(b"\n")?;
    }

    Ok(())
}


fn read_projects_file() -> Result<Projects> {
    let projects_file = dirs::data_path().join("projects");

    if !projects_file.exists() {
        return Ok(Projects::default());
    }

    let mut file = std::fs::File::open(projects_file)?;

    let mut buf = Vec::new();
    file.read_to_end(&mut buf)?;


    let list = buf
        .split(|&x| x == b'\n')
        .filter_map(|chunk| {
            // FIXME: not a good idea to store the OsStr
            let os_str = unsafe { OsStr::from_encoded_bytes_unchecked(chunk) };
            if os_str.is_empty() {
                None
            } else {
                Some(PathBuf::from(os_str))
            }
        })
        .collect();

    Ok(Projects { list })
}
